(in-package :text-utils)

(defun codepoint->int (string)
  (parse-integer string :radix 16))

(defun split-fields (line)
  (cl-ppcre:split ";" line))

(defun valid-line-p (line)
  (and (string-not-empty-p line)
       (not (cl-ppcre:scan "^#" line))))

(defun match-codepoint-p (a b)
  (alexandria:set-equal a
                        (mapcar #'code-char b)
                        :test #'char=))

(defun extract-codepoints (stream)
  "Extract shotcodes from the emoji table file"
  (flet ((readline ()
           (read-line stream nil nil)))
    (let ((res ()))
      (loop for line = (readline) while line do
        (when (valid-line-p line)
          (let* ((fields            (split-fields line))
                 (codepoints-fields (split-words (alexandria:first-elt fields)))
                 (rangesp           (cl-ppcre:scan "\\.\\."
                                                   (first codepoints-fields)))
                 (codepoints-ranges (and rangesp
                                         (mapcar (lambda (a) (codepoint->int a))
                                                 (cl-ppcre:split "\\.\\."
                                                                 (first codepoints-fields)))))
                 (codepoint-list    (and (not rangesp)
                                         (mapcar (lambda (a) (codepoint->int a))
                                                 codepoints-fields))))
            (if (not rangesp)
                (push codepoint-list res)
                (loop for i from (first codepoints-ranges) to (second codepoints-ranges) do
                  (push (list i) res))))))
      res)))

(defun extract-emojis-codepoints (uri)
  (let* ((data (drakma:http-request uri)))
    (with-input-from-string (stream data)
      (extract-codepoints stream))))

(defmacro gen-matcher-sequences ()
  `(defun emoji-sequences-p (chars)
     (find-if (lambda (a) (match-codepoint-p chars a))
              ',(extract-emojis-codepoints "https://unicode.org/Public/emoji/13.1/emoji-sequences.txt"))))

(defmacro gen-matcher-zwj-sequences ()
  `(defun emoji-zwj-sequences-p (chars)
     (find-if (lambda (a) (match-codepoint-p chars a))
              ',(extract-emojis-codepoints "https://unicode.org/Public/emoji/13.1/emoji-zwj-sequences.txt"))))

(defun fetch-max-codepoints-emoji-lengths ()
  (length (reduce (lambda (a b)
                    (if (> (length a) (length b))
                        a
                        b))
                  (text-utils::extract-emojis-codepoints "https://unicode.org/Public/emoji/13.1/emoji-zwj-sequences.txt"))))

(alexandria:define-constant +max-codepoints-emoji-lengths+ 10 :test #'=)

;;;; expansion of (gen-matcher-sequences)

(defun emoji-sequences-p (chars)
  (find-if (lambda (a) (match-codepoint-p chars a))
           '((129501 127999) (129501 127998) (129501 127997) (129501 127996)
             (129501 127995) (129500 127999) (129500 127998) (129500 127997)
             (129500 127996) (129500 127995) (129499 127999) (129499 127998)
             (129499 127997) (129499 127996) (129499 127995) (129498 127999)
             (129498 127998) (129498 127997) (129498 127996) (129498 127995)
             (129497 127999) (129497 127998) (129497 127997) (129497 127996)
             (129497 127995) (129496 127999) (129496 127998) (129496 127997)
             (129496 127996) (129496 127995) (129495 127999) (129495 127998)
             (129495 127997) (129495 127996) (129495 127995) (129494 127999)
             (129494 127998) (129494 127997) (129494 127996) (129494 127995)
             (129493 127999) (129493 127998) (129493 127997) (129493 127996)
             (129493 127995) (129492 127999) (129492 127998) (129492 127997)
             (129492 127996) (129492 127995) (129491 127999) (129491 127998)
             (129491 127997) (129491 127996) (129491 127995) (129490 127999)
             (129490 127998) (129490 127997) (129490 127996) (129490 127995)
             (129489 127999) (129489 127998) (129489 127997) (129489 127996)
             (129489 127995) (129487 127999) (129487 127998) (129487 127997)
             (129487 127996) (129487 127995) (129486 127999) (129486 127998)
             (129486 127997) (129486 127996) (129486 127995) (129485 127999)
             (129485 127998) (129485 127997) (129485 127996) (129485 127995)
             (129467 127999) (129467 127998) (129467 127997) (129467 127996)
             (129467 127995) (129465 127999) (129465 127998) (129465 127997)
             (129465 127996) (129465 127995) (129464 127999) (129464 127998)
             (129464 127997) (129464 127996) (129464 127995) (129462 127999)
             (129462 127998) (129462 127997) (129462 127996) (129462 127995)
             (129461 127999) (129461 127998) (129461 127997) (129461 127996)
             (129461 127995) (129399 127999) (129399 127998) (129399 127997)
             (129399 127996) (129399 127995) (129342 127999) (129342 127998)
             (129342 127997) (129342 127996) (129342 127995) (129341 127999)
             (129341 127998) (129341 127997) (129341 127996) (129341 127995)
             (129337 127999) (129337 127998) (129337 127997) (129337 127996)
             (129337 127995) (129336 127999) (129336 127998) (129336 127997)
             (129336 127996) (129336 127995) (129335 127999) (129335 127998)
             (129335 127997) (129335 127996) (129335 127995) (129334 127999)
             (129334 127998) (129334 127997) (129334 127996) (129334 127995)
             (129333 127999) (129333 127998) (129333 127997) (129333 127996)
             (129333 127995) (129332 127999) (129332 127998) (129332 127997)
             (129332 127996) (129332 127995) (129331 127999) (129331 127998)
             (129331 127997) (129331 127996) (129331 127995) (129330 127999)
             (129330 127998) (129330 127997) (129330 127996) (129330 127995)
             (129329 127999) (129329 127998) (129329 127997) (129329 127996)
             (129329 127995) (129328 127999) (129328 127998) (129328 127997)
             (129328 127996) (129328 127995) (129318 127999) (129318 127998)
             (129318 127997) (129318 127996) (129318 127995) (129311 127999)
             (129311 127998) (129311 127997) (129311 127996) (129311 127995)
             (129310 127999) (129310 127998) (129310 127997) (129310 127996)
             (129310 127995) (129308 127999) (129308 127998) (129308 127997)
             (129308 127996) (129308 127995) (129307 127999) (129307 127998)
             (129307 127997) (129307 127996) (129307 127995) (129306 127999)
             (129306 127998) (129306 127997) (129306 127996) (129306 127995)
             (129305 127999) (129305 127998) (129305 127997) (129305 127996)
             (129305 127995) (129304 127999) (129304 127998) (129304 127997)
             (129304 127996) (129304 127995) (129295 127999) (129295 127998)
             (129295 127997) (129295 127996) (129295 127995) (129292 127999)
             (129292 127998) (129292 127997) (129292 127996) (129292 127995)
             (128716 127999) (128716 127998) (128716 127997) (128716 127996)
             (128716 127995) (128704 127999) (128704 127998) (128704 127997)
             (128704 127996) (128704 127995) (128694 127999) (128694 127998)
             (128694 127997) (128694 127996) (128694 127995) (128693 127999)
             (128693 127998) (128693 127997) (128693 127996) (128693 127995)
             (128692 127999) (128692 127998) (128692 127997) (128692 127996)
             (128692 127995) (128675 127999) (128675 127998) (128675 127997)
             (128675 127996) (128675 127995) (128591 127999) (128591 127998)
             (128591 127997) (128591 127996) (128591 127995) (128590 127999)
             (128590 127998) (128590 127997) (128590 127996) (128590 127995)
             (128589 127999) (128589 127998) (128589 127997) (128589 127996)
             (128589 127995) (128588 127999) (128588 127998) (128588 127997)
             (128588 127996) (128588 127995) (128587 127999) (128587 127998)
             (128587 127997) (128587 127996) (128587 127995) (128583 127999)
             (128583 127998) (128583 127997) (128583 127996) (128583 127995)
             (128582 127999) (128582 127998) (128582 127997) (128582 127996)
             (128582 127995) (128581 127999) (128581 127998) (128581 127997)
             (128581 127996) (128581 127995) (128406 127999) (128406 127998)
             (128406 127997) (128406 127996) (128406 127995) (128405 127999)
             (128405 127998) (128405 127997) (128405 127996) (128405 127995)
             (128400 127999) (128400 127998) (128400 127997) (128400 127996)
             (128400 127995) (128378 127999) (128378 127998) (128378 127997)
             (128378 127996) (128378 127995) (128373 127999) (128373 127998)
             (128373 127997) (128373 127996) (128373 127995) (128372 127999)
             (128372 127998) (128372 127997) (128372 127996) (128372 127995)
             (128170 127999) (128170 127998) (128170 127997) (128170 127996)
             (128170 127995) (128145 127999) (128145 127998) (128145 127997)
             (128145 127996) (128145 127995) (128143 127999) (128143 127998)
             (128143 127997) (128143 127996) (128143 127995) (128135 127999)
             (128135 127998) (128135 127997) (128135 127996) (128135 127995)
             (128134 127999) (128134 127998) (128134 127997) (128134 127996)
             (128134 127995) (128133 127999) (128133 127998) (128133 127997)
             (128133 127996) (128133 127995) (128131 127999) (128131 127998)
             (128131 127997) (128131 127996) (128131 127995) (128130 127999)
             (128130 127998) (128130 127997) (128130 127996) (128130 127995)
             (128129 127999) (128129 127998) (128129 127997) (128129 127996)
             (128129 127995) (128124 127999) (128124 127998) (128124 127997)
             (128124 127996) (128124 127995) (128120 127999) (128120 127998)
             (128120 127997) (128120 127996) (128120 127995) (128119 127999)
             (128119 127998) (128119 127997) (128119 127996) (128119 127995)
             (128118 127999) (128118 127998) (128118 127997) (128118 127996)
             (128118 127995) (128117 127999) (128117 127998) (128117 127997)
             (128117 127996) (128117 127995) (128116 127999) (128116 127998)
             (128116 127997) (128116 127996) (128116 127995) (128115 127999)
             (128115 127998) (128115 127997) (128115 127996) (128115 127995)
             (128114 127999) (128114 127998) (128114 127997) (128114 127996)
             (128114 127995) (128113 127999) (128113 127998) (128113 127997)
             (128113 127996) (128113 127995) (128112 127999) (128112 127998)
             (128112 127997) (128112 127996) (128112 127995) (128110 127999)
             (128110 127998) (128110 127997) (128110 127996) (128110 127995)
             (128109 127999) (128109 127998) (128109 127997) (128109 127996)
             (128109 127995) (128108 127999) (128108 127998) (128108 127997)
             (128108 127996) (128108 127995) (128107 127999) (128107 127998)
             (128107 127997) (128107 127996) (128107 127995) (128105 127999)
             (128105 127998) (128105 127997) (128105 127996) (128105 127995)
             (128104 127999) (128104 127998) (128104 127997) (128104 127996)
             (128104 127995) (128103 127999) (128103 127998) (128103 127997)
             (128103 127996) (128103 127995) (128102 127999) (128102 127998)
             (128102 127997) (128102 127996) (128102 127995) (128080 127999)
             (128080 127998) (128080 127997) (128080 127996) (128080 127995)
             (128079 127999) (128079 127998) (128079 127997) (128079 127996)
             (128079 127995) (128078 127999) (128078 127998) (128078 127997)
             (128078 127996) (128078 127995) (128077 127999) (128077 127998)
             (128077 127997) (128077 127996) (128077 127995) (128076 127999)
             (128076 127998) (128076 127997) (128076 127996) (128076 127995)
             (128075 127999) (128075 127998) (128075 127997) (128075 127996)
             (128075 127995) (128074 127999) (128074 127998) (128074 127997)
             (128074 127996) (128074 127995) (128073 127999) (128073 127998)
             (128073 127997) (128073 127996) (128073 127995) (128072 127999)
             (128072 127998) (128072 127997) (128072 127996) (128072 127995)
             (128071 127999) (128071 127998) (128071 127997) (128071 127996)
             (128071 127995) (128070 127999) (128070 127998) (128070 127997)
             (128070 127996) (128070 127995) (128067 127999) (128067 127998)
             (128067 127997) (128067 127996) (128067 127995) (128066 127999)
             (128066 127998) (128066 127997) (128066 127996) (128066 127995)
             (127948 127999) (127948 127998) (127948 127997) (127948 127996)
             (127948 127995) (127947 127999) (127947 127998) (127947 127997)
             (127947 127996) (127947 127995) (127946 127999) (127946 127998)
             (127946 127997) (127946 127996) (127946 127995) (127943 127999)
             (127943 127998) (127943 127997) (127943 127996) (127943 127995)
             (127940 127999) (127940 127998) (127940 127997) (127940 127996)
             (127940 127995) (127939 127999) (127939 127998) (127939 127997)
             (127939 127996) (127939 127995) (127938 127999) (127938 127998)
             (127938 127997) (127938 127996) (127938 127995) (127877 127999)
             (127877 127998) (127877 127997) (127877 127996) (127877 127995)
             (9997 127999) (9997 127998) (9997 127997) (9997 127996)
             (9997 127995) (9996 127999) (9996 127998) (9996 127997)
             (9996 127996) (9996 127995) (9995 127999) (9995 127998)
             (9995 127997) (9995 127996) (9995 127995) (9994 127999)
             (9994 127998) (9994 127997) (9994 127996) (9994 127995)
             (9977 127999) (9977 127998) (9977 127997) (9977 127996)
             (9977 127995) (9757 127999) (9757 127998) (9757 127997)
             (9757 127996) (9757 127995)
             (127988 917607 917602 917623 917612 917619 917631)
             (127988 917607 917602 917619 917603 917620 917631)
             (127988 917607 917602 917605 917614 917607 917631) (127487 127484)
             (127487 127474) (127487 127462) (127486 127481) (127486 127466)
             (127485 127472) (127484 127480) (127484 127467) (127483 127482)
             (127483 127475) (127483 127470) (127483 127468) (127483 127466)
             (127483 127464) (127483 127462) (127482 127487) (127482 127486)
             (127482 127480) (127482 127475) (127482 127474) (127482 127468)
             (127482 127462) (127481 127487) (127481 127484) (127481 127483)
             (127481 127481) (127481 127479) (127481 127476) (127481 127475)
             (127481 127474) (127481 127473) (127481 127472) (127481 127471)
             (127481 127469) (127481 127468) (127481 127467) (127481 127465)
             (127481 127464) (127481 127462) (127480 127487) (127480 127486)
             (127480 127485) (127480 127483) (127480 127481) (127480 127480)
             (127480 127479) (127480 127476) (127480 127475) (127480 127474)
             (127480 127473) (127480 127472) (127480 127471) (127480 127470)
             (127480 127469) (127480 127468) (127480 127466) (127480 127465)
             (127480 127464) (127480 127463) (127480 127462) (127479 127484)
             (127479 127482) (127479 127480) (127479 127476) (127479 127466)
             (127478 127462) (127477 127486) (127477 127484) (127477 127481)
             (127477 127480) (127477 127479) (127477 127475) (127477 127474)
             (127477 127473) (127477 127472) (127477 127469) (127477 127468)
             (127477 127467) (127477 127466) (127477 127462) (127476 127474)
             (127475 127487) (127475 127482) (127475 127479) (127475 127477)
             (127475 127476) (127475 127473) (127475 127470) (127475 127468)
             (127475 127467) (127475 127466) (127475 127464) (127475 127462)
             (127474 127487) (127474 127486) (127474 127485) (127474 127484)
             (127474 127483) (127474 127482) (127474 127481) (127474 127480)
             (127474 127479) (127474 127478) (127474 127477) (127474 127476)
             (127474 127475) (127474 127474) (127474 127473) (127474 127472)
             (127474 127469) (127474 127468) (127474 127467) (127474 127466)
             (127474 127465) (127474 127464) (127474 127462) (127473 127486)
             (127473 127483) (127473 127482) (127473 127481) (127473 127480)
             (127473 127479) (127473 127472) (127473 127470) (127473 127464)
             (127473 127463) (127473 127462) (127472 127487) (127472 127486)
             (127472 127484) (127472 127479) (127472 127477) (127472 127475)
             (127472 127474) (127472 127470) (127472 127469) (127472 127468)
             (127472 127466) (127471 127477) (127471 127476) (127471 127474)
             (127471 127466) (127470 127481) (127470 127480) (127470 127479)
             (127470 127478) (127470 127476) (127470 127475) (127470 127474)
             (127470 127473) (127470 127466) (127470 127465) (127470 127464)
             (127469 127482) (127469 127481) (127469 127479) (127469 127475)
             (127469 127474) (127469 127472) (127468 127486) (127468 127484)
             (127468 127482) (127468 127481) (127468 127480) (127468 127479)
             (127468 127478) (127468 127477) (127468 127475) (127468 127474)
             (127468 127473) (127468 127470) (127468 127469) (127468 127468)
             (127468 127467) (127468 127466) (127468 127465) (127468 127463)
             (127468 127462) (127467 127479) (127467 127476) (127467 127474)
             (127467 127472) (127467 127471) (127467 127470) (127466 127482)
             (127466 127481) (127466 127480) (127466 127479) (127466 127469)
             (127466 127468) (127466 127466) (127466 127464) (127466 127462)
             (127465 127487) (127465 127476) (127465 127474) (127465 127472)
             (127465 127471) (127465 127468) (127465 127466) (127464 127487)
             (127464 127486) (127464 127485) (127464 127484) (127464 127483)
             (127464 127482) (127464 127479) (127464 127477) (127464 127476)
             (127464 127475) (127464 127474) (127464 127473) (127464 127472)
             (127464 127470) (127464 127469) (127464 127468) (127464 127467)
             (127464 127465) (127464 127464) (127464 127462) (127463 127487)
             (127463 127486) (127463 127484) (127463 127483) (127463 127481)
             (127463 127480) (127463 127479) (127463 127478) (127463 127476)
             (127463 127475) (127463 127474) (127463 127473) (127463 127471)
             (127463 127470) (127463 127469) (127463 127468) (127463 127467)
             (127463 127466) (127463 127465) (127463 127463) (127463 127462)
             (127462 127487) (127462 127485) (127462 127484) (127462 127482)
             (127462 127481) (127462 127480) (127462 127479) (127462 127478)
             (127462 127476) (127462 127474) (127462 127473) (127462 127470)
             (127462 127468) (127462 127467) (127462 127466) (127462 127465)
             (127462 127464) (57 65039 8419) (56 65039 8419) (55 65039 8419)
             (54 65039 8419) (53 65039 8419) (52 65039 8419) (51 65039 8419)
             (50 65039 8419) (49 65039 8419) (48 65039 8419) (42 65039 8419)
             (35 65039 8419) (128755 65039) (128752 65039) (128745 65039)
             (128741 65039) (128740 65039) (128739 65039) (128738 65039)
             (128737 65039) (128736 65039) (128719 65039) (128718 65039)
             (128717 65039) (128715 65039) (128506 65039) (128499 65039)
             (128495 65039) (128488 65039) (128483 65039) (128481 65039)
             (128478 65039) (128477 65039) (128476 65039) (128467 65039)
             (128466 65039) (128465 65039) (128452 65039) (128451 65039)
             (128450 65039) (128444 65039) (128434 65039) (128433 65039)
             (128424 65039) (128421 65039) (128400 65039) (128397 65039)
             (128396 65039) (128395 65039) (128394 65039) (128391 65039)
             (128377 65039) (128376 65039) (128375 65039) (128374 65039)
             (128373 65039) (128372 65039) (128371 65039) (128368 65039)
             (128367 65039) (128330 65039) (128329 65039) (128253 65039)
             (128065 65039) (128063 65039) (127991 65039) (127989 65039)
             (127987 65039) (127967 65039) (127966 65039) (127965 65039)
             (127964 65039) (127963 65039) (127962 65039) (127961 65039)
             (127960 65039) (127959 65039) (127958 65039) (127957 65039)
             (127956 65039) (127950 65039) (127949 65039) (127948 65039)
             (127947 65039) (127903 65039) (127902 65039) (127899 65039)
             (127898 65039) (127897 65039) (127895 65039) (127894 65039)
             (127869 65039) (127798 65039) (127788 65039) (127787 65039)
             (127786 65039) (127785 65039) (127784 65039) (127783 65039)
             (127782 65039) (127781 65039) (127780 65039) (127777 65039)
             (127543 65039) (127490 65039) (127359 65039) (127358 65039)
             (127345 65039) (127344 65039) (12953 65039) (12951 65039)
             (12349 65039) (12336 65039) (11015 65039) (11014 65039)
             (11013 65039) (10549 65039) (10548 65039) (10145 65039)
             (10084 65039) (10083 65039) (10055 65039) (10052 65039)
             (10036 65039) (10035 65039) (10017 65039) (10013 65039)
             (10006 65039) (10004 65039) (10002 65039) (9999 65039)
             (9997 65039) (9996 65039) (9993 65039) (9992 65039) (9986 65039)
             (9977 65039) (9976 65039) (9975 65039) (9972 65039) (9969 65039)
             (9968 65039) (9961 65039) (9939 65039) (9937 65039) (9935 65039)
             (9928 65039) (9905 65039) (9904 65039) (9895 65039) (9888 65039)
             (9884 65039) (9883 65039) (9881 65039) (9879 65039) (9878 65039)
             (9877 65039) (9876 65039) (9874 65039) (9854 65039) (9851 65039)
             (9832 65039) (9830 65039) (9829 65039) (9827 65039) (9824 65039)
             (9823 65039) (9794 65039) (9792 65039) (9786 65039) (9785 65039)
             (9784 65039) (9775 65039) (9774 65039) (9770 65039) (9766 65039)
             (9763 65039) (9762 65039) (9760 65039) (9757 65039) (9752 65039)
             (9745 65039) (9742 65039) (9732 65039) (9731 65039) (9730 65039)
             (9729 65039) (9728 65039) (9724 65039) (9723 65039) (9664 65039)
             (9654 65039) (9643 65039) (9642 65039) (9410 65039) (9210 65039)
             (9209 65039) (9208 65039) (9202 65039) (9201 65039) (9199 65039)
             (9198 65039) (9197 65039) (9167 65039) (9000 65039) (8618 65039)
             (8617 65039) (8601 65039) (8600 65039) (8599 65039) (8598 65039)
             (8597 65039) (8596 65039) (8505 65039) (8482 65039) (8265 65039)
             (8252 65039) (174 65039) (169 65039) (129750) (129749) (129748)
             (129747) (129746) (129745) (129744) (129730) (129729) (129728)
             (129718) (129717) (129716) (129715) (129714) (129713) (129712)
             (129704) (129703) (129702) (129701) (129700) (129699) (129698)
             (129697) (129696) (129695) (129694) (129693) (129692) (129691)
             (129690) (129689) (129688) (129687) (129686) (129685) (129684)
             (129683) (129682) (129681) (129680) (129670) (129669) (129668)
             (129667) (129666) (129665) (129664) (129658) (129657) (129656)
             (129652) (129651) (129650) (129649) (129648) (129535) (129534)
             (129533) (129532) (129531) (129530) (129529) (129528) (129527)
             (129526) (129525) (129524) (129523) (129522) (129521) (129520)
             (129519) (129518) (129517) (129516) (129515) (129514) (129513)
             (129512) (129511) (129510) (129509) (129508) (129507) (129506)
             (129505) (129504) (129503) (129502) (129501) (129500) (129499)
             (129498) (129497) (129496) (129495) (129494) (129493) (129492)
             (129491) (129490) (129489) (129488) (129487) (129486) (129485)
             (129483) (129482) (129481) (129480) (129479) (129478) (129477)
             (129476) (129475) (129474) (129473) (129472) (129471) (129470)
             (129469) (129468) (129467) (129466) (129465) (129464) (129463)
             (129462) (129461) (129460) (129459) (129458) (129457) (129456)
             (129455) (129454) (129453) (129452) (129451) (129450) (129449)
             (129448) (129447) (129446) (129445) (129444) (129443) (129442)
             (129441) (129440) (129439) (129438) (129437) (129436) (129435)
             (129434) (129433) (129432) (129431) (129430) (129429) (129428)
             (129427) (129426) (129425) (129424) (129423) (129422) (129421)
             (129420) (129419) (129418) (129417) (129416) (129415) (129414)
             (129413) (129412) (129411) (129410) (129409) (129408) (129407)
             (129406) (129405) (129404) (129403) (129402) (129400) (129399)
             (129398) (129397) (129396) (129395) (129394) (129393) (129392)
             (129391) (129390) (129389) (129388) (129387) (129386) (129385)
             (129384) (129383) (129382) (129381) (129380) (129379) (129378)
             (129377) (129376) (129375) (129374) (129373) (129372) (129371)
             (129370) (129369) (129368) (129367) (129366) (129365) (129364)
             (129363) (129362) (129361) (129360) (129359) (129358) (129357)
             (129356) (129355) (129354) (129353) (129352) (129351) (129349)
             (129348) (129347) (129346) (129345) (129344) (129343) (129342)
             (129341) (129340) (129338) (129337) (129336) (129335) (129334)
             (129333) (129332) (129331) (129330) (129329) (129328) (129327)
             (129326) (129325) (129324) (129323) (129322) (129321) (129320)
             (129319) (129318) (129317) (129316) (129315) (129314) (129313)
             (129312) (129311) (129310) (129309) (129308) (129307) (129306)
             (129305) (129304) (129303) (129302) (129301) (129300) (129299)
             (129298) (129297) (129296) (129295) (129294) (129293) (129292)
             (129003) (129002) (129001) (129000) (128999) (128998) (128997)
             (128996) (128995) (128994) (128993) (128992) (128764) (128763)
             (128762) (128761) (128760) (128759) (128758) (128757) (128756)
             (128748) (128747) (128727) (128726) (128725) (128722) (128721)
             (128720) (128716) (128709) (128708) (128707) (128706) (128705)
             (128704) (128703) (128702) (128701) (128700) (128699) (128698)
             (128697) (128696) (128695) (128694) (128693) (128692) (128691)
             (128690) (128689) (128688) (128687) (128686) (128685) (128684)
             (128683) (128682) (128681) (128680) (128679) (128678) (128677)
             (128676) (128675) (128674) (128673) (128672) (128671) (128670)
             (128669) (128668) (128667) (128666) (128665) (128664) (128663)
             (128662) (128661) (128660) (128659) (128658) (128657) (128656)
             (128655) (128654) (128653) (128652) (128651) (128650) (128649)
             (128648) (128647) (128646) (128645) (128644) (128643) (128642)
             (128641) (128640) (128591) (128590) (128589) (128588) (128587)
             (128586) (128585) (128584) (128583) (128582) (128581) (128580)
             (128579) (128578) (128577) (128576) (128575) (128574) (128573)
             (128572) (128571) (128570) (128569) (128568) (128567) (128566)
             (128565) (128564) (128563) (128562) (128561) (128560) (128559)
             (128558) (128557) (128556) (128555) (128554) (128553) (128552)
             (128551) (128550) (128549) (128548) (128547) (128546) (128545)
             (128544) (128543) (128542) (128541) (128540) (128539) (128538)
             (128537) (128536) (128535) (128534) (128533) (128532) (128531)
             (128530) (128529) (128528) (128527) (128526) (128525) (128524)
             (128523) (128522) (128521) (128520) (128519) (128518) (128517)
             (128516) (128515) (128514) (128513) (128512) (128511) (128510)
             (128509) (128508) (128507) (128420) (128406) (128405) (128378)
             (128359) (128358) (128357) (128356) (128355) (128354) (128353)
             (128352) (128351) (128350) (128349) (128348) (128347) (128346)
             (128345) (128344) (128343) (128342) (128341) (128340) (128339)
             (128338) (128337) (128336) (128334) (128333) (128332) (128331)
             (128317) (128316) (128315) (128314) (128313) (128312) (128311)
             (128310) (128309) (128308) (128307) (128306) (128305) (128304)
             (128303) (128302) (128301) (128300) (128299) (128298) (128297)
             (128296) (128295) (128294) (128293) (128292) (128291) (128290)
             (128289) (128288) (128287) (128286) (128285) (128284) (128283)
             (128282) (128281) (128280) (128279) (128278) (128277) (128276)
             (128275) (128274) (128273) (128272) (128271) (128270) (128269)
             (128268) (128267) (128266) (128265) (128264) (128263) (128262)
             (128261) (128260) (128259) (128258) (128257) (128256) (128255)
             (128252) (128251) (128250) (128249) (128248) (128247) (128246)
             (128245) (128244) (128243) (128242) (128241) (128240) (128239)
             (128238) (128237) (128236) (128235) (128234) (128233) (128232)
             (128231) (128230) (128229) (128228) (128227) (128226) (128225)
             (128224) (128223) (128222) (128221) (128220) (128219) (128218)
             (128217) (128216) (128215) (128214) (128213) (128212) (128211)
             (128210) (128209) (128208) (128207) (128206) (128205) (128204)
             (128203) (128202) (128201) (128200) (128199) (128198) (128197)
             (128196) (128195) (128194) (128193) (128192) (128191) (128190)
             (128189) (128188) (128187) (128186) (128185) (128184) (128183)
             (128182) (128181) (128180) (128179) (128178) (128177) (128176)
             (128175) (128174) (128173) (128172) (128171) (128170) (128169)
             (128168) (128167) (128166) (128165) (128164) (128163) (128162)
             (128161) (128160) (128159) (128158) (128157) (128156) (128155)
             (128154) (128153) (128152) (128151) (128150) (128149) (128148)
             (128147) (128146) (128145) (128144) (128143) (128142) (128141)
             (128140) (128139) (128138) (128137) (128136) (128135) (128134)
             (128133) (128132) (128131) (128130) (128129) (128128) (128127)
             (128126) (128125) (128124) (128123) (128122) (128121) (128120)
             (128119) (128118) (128117) (128116) (128115) (128114) (128113)
             (128112) (128111) (128110) (128109) (128108) (128107) (128106)
             (128105) (128104) (128103) (128102) (128101) (128100) (128099)
             (128098) (128097) (128096) (128095) (128094) (128093) (128092)
             (128091) (128090) (128089) (128088) (128087) (128086) (128085)
             (128084) (128083) (128082) (128081) (128080) (128079) (128078)
             (128077) (128076) (128075) (128074) (128073) (128072) (128071)
             (128070) (128069) (128068) (128067) (128066) (128064) (128062)
             (128061) (128060) (128059) (128058) (128057) (128056) (128055)
             (128054) (128053) (128052) (128051) (128050) (128049) (128048)
             (128047) (128046) (128045) (128044) (128043) (128042) (128041)
             (128040) (128039) (128038) (128037) (128036) (128035) (128034)
             (128033) (128032) (128031) (128030) (128029) (128028) (128027)
             (128026) (128025) (128024) (128023) (128022) (128021) (128020)
             (128019) (128018) (128017) (128016) (128015) (128014) (128013)
             (128012) (128011) (128010) (128009) (128008) (128007) (128006)
             (128005) (128004) (128003) (128002) (128001) (128000) (127999)
             (127998) (127997) (127996) (127995) (127994) (127993) (127992)
             (127988) (127984) (127983) (127982) (127981) (127980) (127979)
             (127978) (127977) (127976) (127975) (127974) (127973) (127972)
             (127971) (127970) (127969) (127968) (127955) (127954) (127953)
             (127952) (127951) (127946) (127945) (127944) (127943) (127942)
             (127941) (127940) (127939) (127938) (127937) (127936) (127935)
             (127934) (127933) (127932) (127931) (127930) (127929) (127928)
             (127927) (127926) (127925) (127924) (127923) (127922) (127921)
             (127920) (127919) (127918) (127917) (127916) (127915) (127914)
             (127913) (127912) (127911) (127910) (127909) (127908) (127907)
             (127906) (127905) (127904) (127891) (127890) (127889) (127888)
             (127887) (127886) (127885) (127884) (127883) (127882) (127881)
             (127880) (127879) (127878) (127877) (127876) (127875) (127874)
             (127873) (127872) (127871) (127870) (127868) (127867) (127866)
             (127865) (127864) (127863) (127862) (127861) (127860) (127859)
             (127858) (127857) (127856) (127855) (127854) (127853) (127852)
             (127851) (127850) (127849) (127848) (127847) (127846) (127845)
             (127844) (127843) (127842) (127841) (127840) (127839) (127838)
             (127837) (127836) (127835) (127834) (127833) (127832) (127831)
             (127830) (127829) (127828) (127827) (127826) (127825) (127824)
             (127823) (127822) (127821) (127820) (127819) (127818) (127817)
             (127816) (127815) (127814) (127813) (127812) (127811) (127810)
             (127809) (127808) (127807) (127806) (127805) (127804) (127803)
             (127802) (127801) (127800) (127799) (127797) (127796) (127795)
             (127794) (127793) (127792) (127791) (127790) (127789) (127776)
             (127775) (127774) (127773) (127772) (127771) (127770) (127769)
             (127768) (127767) (127766) (127765) (127764) (127763) (127762)
             (127761) (127760) (127759) (127758) (127757) (127756) (127755)
             (127754) (127753) (127752) (127751) (127750) (127749) (127748)
             (127747) (127746) (127745) (127744) (127569) (127568) (127546)
             (127545) (127544) (127542) (127541) (127540) (127539) (127538)
             (127535) (127514) (127489) (127386) (127385) (127384) (127383)
             (127382) (127381) (127380) (127379) (127378) (127377) (127374)
             (127183) (126980) (11093) (11088) (11036) (11035) (10175) (10160)
             (10135) (10134) (10133) (10071) (10069) (10068) (10067) (10062)
             (10060) (10024) (9995) (9994) (9989) (9981) (9978) (9973) (9971)
             (9970) (9962) (9940) (9934) (9925) (9924) (9918) (9917) (9899)
             (9898) (9889) (9875) (9855) (9811) (9810) (9809) (9808) (9807)
             (9806) (9805) (9804) (9803) (9802) (9801) (9800) (9749) (9748)
             (9726) (9725) (9203) (9200) (9196) (9195) (9194) (9193) (8987)
             (8986))))

;;;; expansion of (gen-matcher-zwj-sequences)

(defun emoji-zwj-sequences-p (chars)
  (find-if (lambda (a) (match-codepoint-p chars a))
           '((129489 8205 127876) (128566 8205 127787 65039)
             (128565 8205 128171) (128558 8205 128168)
             (128065 65039 8205 128488 65039) (128059 8205 10052 65039)
             (128021 8205 129466) (128008 8205 11035) (127988 8205 9760 65039)
             (127987 65039 8205 127752) (127987 65039 8205 9895 65039)
             (10084 65039 8205 129657) (10084 65039 8205 128293)
             (129489 127999 8205 129459) (129489 127999 8205 129458)
             (129489 127999 8205 129457) (129489 127999 8205 129456)
             (129489 127998 8205 129459) (129489 127998 8205 129458)
             (129489 127998 8205 129457) (129489 127998 8205 129456)
             (129489 127997 8205 129459) (129489 127997 8205 129458)
             (129489 127997 8205 129457) (129489 127997 8205 129456)
             (129489 127996 8205 129459) (129489 127996 8205 129458)
             (129489 127996 8205 129457) (129489 127996 8205 129456)
             (129489 127995 8205 129459) (129489 127995 8205 129458)
             (129489 127995 8205 129457) (129489 127995 8205 129456)
             (129489 8205 129459) (129489 8205 129458) (129489 8205 129457)
             (129489 8205 129456) (128105 127999 8205 129459)
             (128105 127999 8205 129458) (128105 127999 8205 129457)
             (128105 127999 8205 129456) (128105 127998 8205 129459)
             (128105 127998 8205 129458) (128105 127998 8205 129457)
             (128105 127998 8205 129456) (128105 127997 8205 129459)
             (128105 127997 8205 129458) (128105 127997 8205 129457)
             (128105 127997 8205 129456) (128105 127996 8205 129459)
             (128105 127996 8205 129458) (128105 127996 8205 129457)
             (128105 127996 8205 129456) (128105 127995 8205 129459)
             (128105 127995 8205 129458) (128105 127995 8205 129457)
             (128105 127995 8205 129456) (128105 8205 129459)
             (128105 8205 129458) (128105 8205 129457) (128105 8205 129456)
             (128104 127999 8205 129459) (128104 127999 8205 129458)
             (128104 127999 8205 129457) (128104 127999 8205 129456)
             (128104 127998 8205 129459) (128104 127998 8205 129458)
             (128104 127998 8205 129457) (128104 127998 8205 129456)
             (128104 127997 8205 129459) (128104 127997 8205 129458)
             (128104 127997 8205 129457) (128104 127997 8205 129456)
             (128104 127996 8205 129459) (128104 127996 8205 129458)
             (128104 127996 8205 129457) (128104 127996 8205 129456)
             (128104 127995 8205 129459) (128104 127995 8205 129458)
             (128104 127995 8205 129457) (128104 127995 8205 129456)
             (128104 8205 129459) (128104 8205 129458) (128104 8205 129457)
             (128104 8205 129456) (129503 8205 9794 65039)
             (129503 8205 9792 65039) (129502 8205 9794 65039)
             (129502 8205 9792 65039) (129501 127999 8205 9794 65039)
             (129501 127999 8205 9792 65039) (129501 127998 8205 9794 65039)
             (129501 127998 8205 9792 65039) (129501 127997 8205 9794 65039)
             (129501 127997 8205 9792 65039) (129501 127996 8205 9794 65039)
             (129501 127996 8205 9792 65039) (129501 127995 8205 9794 65039)
             (129501 127995 8205 9792 65039) (129501 8205 9794 65039)
             (129501 8205 9792 65039) (129500 127999 8205 9794 65039)
             (129500 127999 8205 9792 65039) (129500 127998 8205 9794 65039)
             (129500 127998 8205 9792 65039) (129500 127997 8205 9794 65039)
             (129500 127997 8205 9792 65039) (129500 127996 8205 9794 65039)
             (129500 127996 8205 9792 65039) (129500 127995 8205 9794 65039)
             (129500 127995 8205 9792 65039) (129500 8205 9794 65039)
             (129500 8205 9792 65039) (129499 127999 8205 9794 65039)
             (129499 127999 8205 9792 65039) (129499 127998 8205 9794 65039)
             (129499 127998 8205 9792 65039) (129499 127997 8205 9794 65039)
             (129499 127997 8205 9792 65039) (129499 127996 8205 9794 65039)
             (129499 127996 8205 9792 65039) (129499 127995 8205 9794 65039)
             (129499 127995 8205 9792 65039) (129499 8205 9794 65039)
             (129499 8205 9792 65039) (129498 127999 8205 9794 65039)
             (129498 127999 8205 9792 65039) (129498 127998 8205 9794 65039)
             (129498 127998 8205 9792 65039) (129498 127997 8205 9794 65039)
             (129498 127997 8205 9792 65039) (129498 127996 8205 9794 65039)
             (129498 127996 8205 9792 65039) (129498 127995 8205 9794 65039)
             (129498 127995 8205 9792 65039) (129498 8205 9794 65039)
             (129498 8205 9792 65039) (129497 127999 8205 9794 65039)
             (129497 127999 8205 9792 65039) (129497 127998 8205 9794 65039)
             (129497 127998 8205 9792 65039) (129497 127997 8205 9794 65039)
             (129497 127997 8205 9792 65039) (129497 127996 8205 9794 65039)
             (129497 127996 8205 9792 65039) (129497 127995 8205 9794 65039)
             (129497 127995 8205 9792 65039) (129497 8205 9794 65039)
             (129497 8205 9792 65039) (129496 127999 8205 9794 65039)
             (129496 127999 8205 9792 65039) (129496 127998 8205 9794 65039)
             (129496 127998 8205 9792 65039) (129496 127997 8205 9794 65039)
             (129496 127997 8205 9792 65039) (129496 127996 8205 9794 65039)
             (129496 127996 8205 9792 65039) (129496 127995 8205 9794 65039)
             (129496 127995 8205 9792 65039) (129496 8205 9794 65039)
             (129496 8205 9792 65039) (129495 127999 8205 9794 65039)
             (129495 127999 8205 9792 65039) (129495 127998 8205 9794 65039)
             (129495 127998 8205 9792 65039) (129495 127997 8205 9794 65039)
             (129495 127997 8205 9792 65039) (129495 127996 8205 9794 65039)
             (129495 127996 8205 9792 65039) (129495 127995 8205 9794 65039)
             (129495 127995 8205 9792 65039) (129495 8205 9794 65039)
             (129495 8205 9792 65039) (129494 127999 8205 9794 65039)
             (129494 127999 8205 9792 65039) (129494 127998 8205 9794 65039)
             (129494 127998 8205 9792 65039) (129494 127997 8205 9794 65039)
             (129494 127997 8205 9792 65039) (129494 127996 8205 9794 65039)
             (129494 127996 8205 9792 65039) (129494 127995 8205 9794 65039)
             (129494 127995 8205 9792 65039) (129494 8205 9794 65039)
             (129494 8205 9792 65039) (129492 127999 8205 9794 65039)
             (129492 127999 8205 9792 65039) (129492 127998 8205 9794 65039)
             (129492 127998 8205 9792 65039) (129492 127997 8205 9794 65039)
             (129492 127997 8205 9792 65039) (129492 127996 8205 9794 65039)
             (129492 127996 8205 9792 65039) (129492 127995 8205 9794 65039)
             (129492 127995 8205 9792 65039) (129492 8205 9794 65039)
             (129492 8205 9792 65039) (129487 127999 8205 9794 65039)
             (129487 127999 8205 9792 65039) (129487 127998 8205 9794 65039)
             (129487 127998 8205 9792 65039) (129487 127997 8205 9794 65039)
             (129487 127997 8205 9792 65039) (129487 127996 8205 9794 65039)
             (129487 127996 8205 9792 65039) (129487 127995 8205 9794 65039)
             (129487 127995 8205 9792 65039) (129487 8205 9794 65039)
             (129487 8205 9792 65039) (129486 127999 8205 9794 65039)
             (129486 127999 8205 9792 65039) (129486 127998 8205 9794 65039)
             (129486 127998 8205 9792 65039) (129486 127997 8205 9794 65039)
             (129486 127997 8205 9792 65039) (129486 127996 8205 9794 65039)
             (129486 127996 8205 9792 65039) (129486 127995 8205 9794 65039)
             (129486 127995 8205 9792 65039) (129486 8205 9794 65039)
             (129486 8205 9792 65039) (129485 127999 8205 9794 65039)
             (129485 127999 8205 9792 65039) (129485 127998 8205 9794 65039)
             (129485 127998 8205 9792 65039) (129485 127997 8205 9794 65039)
             (129485 127997 8205 9792 65039) (129485 127996 8205 9794 65039)
             (129485 127996 8205 9792 65039) (129485 127995 8205 9794 65039)
             (129485 127995 8205 9792 65039) (129485 8205 9794 65039)
             (129485 8205 9792 65039) (129465 127999 8205 9794 65039)
             (129465 127999 8205 9792 65039) (129465 127998 8205 9794 65039)
             (129465 127998 8205 9792 65039) (129465 127997 8205 9794 65039)
             (129465 127997 8205 9792 65039) (129465 127996 8205 9794 65039)
             (129465 127996 8205 9792 65039) (129465 127995 8205 9794 65039)
             (129465 127995 8205 9792 65039) (129465 8205 9794 65039)
             (129465 8205 9792 65039) (129464 127999 8205 9794 65039)
             (129464 127999 8205 9792 65039) (129464 127998 8205 9794 65039)
             (129464 127998 8205 9792 65039) (129464 127997 8205 9794 65039)
             (129464 127997 8205 9792 65039) (129464 127996 8205 9794 65039)
             (129464 127996 8205 9792 65039) (129464 127995 8205 9794 65039)
             (129464 127995 8205 9792 65039) (129464 8205 9794 65039)
             (129464 8205 9792 65039) (129342 127999 8205 9794 65039)
             (129342 127999 8205 9792 65039) (129342 127998 8205 9794 65039)
             (129342 127998 8205 9792 65039) (129342 127997 8205 9794 65039)
             (129342 127997 8205 9792 65039) (129342 127996 8205 9794 65039)
             (129342 127996 8205 9792 65039) (129342 127995 8205 9794 65039)
             (129342 127995 8205 9792 65039) (129342 8205 9794 65039)
             (129342 8205 9792 65039) (129341 127999 8205 9794 65039)
             (129341 127999 8205 9792 65039) (129341 127998 8205 9794 65039)
             (129341 127998 8205 9792 65039) (129341 127997 8205 9794 65039)
             (129341 127997 8205 9792 65039) (129341 127996 8205 9794 65039)
             (129341 127996 8205 9792 65039) (129341 127995 8205 9794 65039)
             (129341 127995 8205 9792 65039) (129341 8205 9794 65039)
             (129341 8205 9792 65039) (129340 8205 9794 65039)
             (129340 8205 9792 65039) (129337 127999 8205 9794 65039)
             (129337 127999 8205 9792 65039) (129337 127998 8205 9794 65039)
             (129337 127998 8205 9792 65039) (129337 127997 8205 9794 65039)
             (129337 127997 8205 9792 65039) (129337 127996 8205 9794 65039)
             (129337 127996 8205 9792 65039) (129337 127995 8205 9794 65039)
             (129337 127995 8205 9792 65039) (129337 8205 9794 65039)
             (129337 8205 9792 65039) (129336 127999 8205 9794 65039)
             (129336 127999 8205 9792 65039) (129336 127998 8205 9794 65039)
             (129336 127998 8205 9792 65039) (129336 127997 8205 9794 65039)
             (129336 127997 8205 9792 65039) (129336 127996 8205 9794 65039)
             (129336 127996 8205 9792 65039) (129336 127995 8205 9794 65039)
             (129336 127995 8205 9792 65039) (129336 8205 9794 65039)
             (129336 8205 9792 65039) (129335 127999 8205 9794 65039)
             (129335 127999 8205 9792 65039) (129335 127998 8205 9794 65039)
             (129335 127998 8205 9792 65039) (129335 127997 8205 9794 65039)
             (129335 127997 8205 9792 65039) (129335 127996 8205 9794 65039)
             (129335 127996 8205 9792 65039) (129335 127995 8205 9794 65039)
             (129335 127995 8205 9792 65039) (129335 8205 9794 65039)
             (129335 8205 9792 65039) (129333 127999 8205 9794 65039)
             (129333 127999 8205 9792 65039) (129333 127998 8205 9794 65039)
             (129333 127998 8205 9792 65039) (129333 127997 8205 9794 65039)
             (129333 127997 8205 9792 65039) (129333 127996 8205 9794 65039)
             (129333 127996 8205 9792 65039) (129333 127995 8205 9794 65039)
             (129333 127995 8205 9792 65039) (129333 8205 9794 65039)
             (129333 8205 9792 65039) (129318 127999 8205 9794 65039)
             (129318 127999 8205 9792 65039) (129318 127998 8205 9794 65039)
             (129318 127998 8205 9792 65039) (129318 127997 8205 9794 65039)
             (129318 127997 8205 9792 65039) (129318 127996 8205 9794 65039)
             (129318 127996 8205 9792 65039) (129318 127995 8205 9794 65039)
             (129318 127995 8205 9792 65039) (129318 8205 9794 65039)
             (129318 8205 9792 65039) (128694 127999 8205 9794 65039)
             (128694 127999 8205 9792 65039) (128694 127998 8205 9794 65039)
             (128694 127998 8205 9792 65039) (128694 127997 8205 9794 65039)
             (128694 127997 8205 9792 65039) (128694 127996 8205 9794 65039)
             (128694 127996 8205 9792 65039) (128694 127995 8205 9794 65039)
             (128694 127995 8205 9792 65039) (128694 8205 9794 65039)
             (128694 8205 9792 65039) (128693 127999 8205 9794 65039)
             (128693 127999 8205 9792 65039) (128693 127998 8205 9794 65039)
             (128693 127998 8205 9792 65039) (128693 127997 8205 9794 65039)
             (128693 127997 8205 9792 65039) (128693 127996 8205 9794 65039)
             (128693 127996 8205 9792 65039) (128693 127995 8205 9794 65039)
             (128693 127995 8205 9792 65039) (128693 8205 9794 65039)
             (128693 8205 9792 65039) (128692 127999 8205 9794 65039)
             (128692 127999 8205 9792 65039) (128692 127998 8205 9794 65039)
             (128692 127998 8205 9792 65039) (128692 127997 8205 9794 65039)
             (128692 127997 8205 9792 65039) (128692 127996 8205 9794 65039)
             (128692 127996 8205 9792 65039) (128692 127995 8205 9794 65039)
             (128692 127995 8205 9792 65039) (128692 8205 9794 65039)
             (128692 8205 9792 65039) (128675 127999 8205 9794 65039)
             (128675 127999 8205 9792 65039) (128675 127998 8205 9794 65039)
             (128675 127998 8205 9792 65039) (128675 127997 8205 9794 65039)
             (128675 127997 8205 9792 65039) (128675 127996 8205 9794 65039)
             (128675 127996 8205 9792 65039) (128675 127995 8205 9794 65039)
             (128675 127995 8205 9792 65039) (128675 8205 9794 65039)
             (128675 8205 9792 65039) (128590 127999 8205 9794 65039)
             (128590 127999 8205 9792 65039) (128590 127998 8205 9794 65039)
             (128590 127998 8205 9792 65039) (128590 127997 8205 9794 65039)
             (128590 127997 8205 9792 65039) (128590 127996 8205 9794 65039)
             (128590 127996 8205 9792 65039) (128590 127995 8205 9794 65039)
             (128590 127995 8205 9792 65039) (128590 8205 9794 65039)
             (128590 8205 9792 65039) (128589 127999 8205 9794 65039)
             (128589 127999 8205 9792 65039) (128589 127998 8205 9794 65039)
             (128589 127998 8205 9792 65039) (128589 127997 8205 9794 65039)
             (128589 127997 8205 9792 65039) (128589 127996 8205 9794 65039)
             (128589 127996 8205 9792 65039) (128589 127995 8205 9794 65039)
             (128589 127995 8205 9792 65039) (128589 8205 9794 65039)
             (128589 8205 9792 65039) (128587 127999 8205 9794 65039)
             (128587 127999 8205 9792 65039) (128587 127998 8205 9794 65039)
             (128587 127998 8205 9792 65039) (128587 127997 8205 9794 65039)
             (128587 127997 8205 9792 65039) (128587 127996 8205 9794 65039)
             (128587 127996 8205 9792 65039) (128587 127995 8205 9794 65039)
             (128587 127995 8205 9792 65039) (128587 8205 9794 65039)
             (128587 8205 9792 65039) (128583 127999 8205 9794 65039)
             (128583 127999 8205 9792 65039) (128583 127998 8205 9794 65039)
             (128583 127998 8205 9792 65039) (128583 127997 8205 9794 65039)
             (128583 127997 8205 9792 65039) (128583 127996 8205 9794 65039)
             (128583 127996 8205 9792 65039) (128583 127995 8205 9794 65039)
             (128583 127995 8205 9792 65039) (128583 8205 9794 65039)
             (128583 8205 9792 65039) (128582 127999 8205 9794 65039)
             (128582 127999 8205 9792 65039) (128582 127998 8205 9794 65039)
             (128582 127998 8205 9792 65039) (128582 127997 8205 9794 65039)
             (128582 127997 8205 9792 65039) (128582 127996 8205 9794 65039)
             (128582 127996 8205 9792 65039) (128582 127995 8205 9794 65039)
             (128582 127995 8205 9792 65039) (128582 8205 9794 65039)
             (128582 8205 9792 65039) (128581 127999 8205 9794 65039)
             (128581 127999 8205 9792 65039) (128581 127998 8205 9794 65039)
             (128581 127998 8205 9792 65039) (128581 127997 8205 9794 65039)
             (128581 127997 8205 9792 65039) (128581 127996 8205 9794 65039)
             (128581 127996 8205 9792 65039) (128581 127995 8205 9794 65039)
             (128581 127995 8205 9792 65039) (128581 8205 9794 65039)
             (128581 8205 9792 65039) (128373 65039 8205 9794 65039)
             (128373 65039 8205 9792 65039) (128373 127999 8205 9794 65039)
             (128373 127999 8205 9792 65039) (128373 127998 8205 9794 65039)
             (128373 127998 8205 9792 65039) (128373 127997 8205 9794 65039)
             (128373 127997 8205 9792 65039) (128373 127996 8205 9794 65039)
             (128373 127996 8205 9792 65039) (128373 127995 8205 9794 65039)
             (128373 127995 8205 9792 65039) (128135 127999 8205 9794 65039)
             (128135 127999 8205 9792 65039) (128135 127998 8205 9794 65039)
             (128135 127998 8205 9792 65039) (128135 127997 8205 9794 65039)
             (128135 127997 8205 9792 65039) (128135 127996 8205 9794 65039)
             (128135 127996 8205 9792 65039) (128135 127995 8205 9794 65039)
             (128135 127995 8205 9792 65039) (128135 8205 9794 65039)
             (128135 8205 9792 65039) (128134 127999 8205 9794 65039)
             (128134 127999 8205 9792 65039) (128134 127998 8205 9794 65039)
             (128134 127998 8205 9792 65039) (128134 127997 8205 9794 65039)
             (128134 127997 8205 9792 65039) (128134 127996 8205 9794 65039)
             (128134 127996 8205 9792 65039) (128134 127995 8205 9794 65039)
             (128134 127995 8205 9792 65039) (128134 8205 9794 65039)
             (128134 8205 9792 65039) (128130 127999 8205 9794 65039)
             (128130 127999 8205 9792 65039) (128130 127998 8205 9794 65039)
             (128130 127998 8205 9792 65039) (128130 127997 8205 9794 65039)
             (128130 127997 8205 9792 65039) (128130 127996 8205 9794 65039)
             (128130 127996 8205 9792 65039) (128130 127995 8205 9794 65039)
             (128130 127995 8205 9792 65039) (128130 8205 9794 65039)
             (128130 8205 9792 65039) (128129 127999 8205 9794 65039)
             (128129 127999 8205 9792 65039) (128129 127998 8205 9794 65039)
             (128129 127998 8205 9792 65039) (128129 127997 8205 9794 65039)
             (128129 127997 8205 9792 65039) (128129 127996 8205 9794 65039)
             (128129 127996 8205 9792 65039) (128129 127995 8205 9794 65039)
             (128129 127995 8205 9792 65039) (128129 8205 9794 65039)
             (128129 8205 9792 65039) (128119 127999 8205 9794 65039)
             (128119 127999 8205 9792 65039) (128119 127998 8205 9794 65039)
             (128119 127998 8205 9792 65039) (128119 127997 8205 9794 65039)
             (128119 127997 8205 9792 65039) (128119 127996 8205 9794 65039)
             (128119 127996 8205 9792 65039) (128119 127995 8205 9794 65039)
             (128119 127995 8205 9792 65039) (128119 8205 9794 65039)
             (128119 8205 9792 65039) (128115 127999 8205 9794 65039)
             (128115 127999 8205 9792 65039) (128115 127998 8205 9794 65039)
             (128115 127998 8205 9792 65039) (128115 127997 8205 9794 65039)
             (128115 127997 8205 9792 65039) (128115 127996 8205 9794 65039)
             (128115 127996 8205 9792 65039) (128115 127995 8205 9794 65039)
             (128115 127995 8205 9792 65039) (128115 8205 9794 65039)
             (128115 8205 9792 65039) (128113 127999 8205 9794 65039)
             (128113 127999 8205 9792 65039) (128113 127998 8205 9794 65039)
             (128113 127998 8205 9792 65039) (128113 127997 8205 9794 65039)
             (128113 127997 8205 9792 65039) (128113 127996 8205 9794 65039)
             (128113 127996 8205 9792 65039) (128113 127995 8205 9794 65039)
             (128113 127995 8205 9792 65039) (128113 8205 9794 65039)
             (128113 8205 9792 65039) (128112 127999 8205 9794 65039)
             (128112 127999 8205 9792 65039) (128112 127998 8205 9794 65039)
             (128112 127998 8205 9792 65039) (128112 127997 8205 9794 65039)
             (128112 127997 8205 9792 65039) (128112 127996 8205 9794 65039)
             (128112 127996 8205 9792 65039) (128112 127995 8205 9794 65039)
             (128112 127995 8205 9792 65039) (128112 8205 9794 65039)
             (128112 8205 9792 65039) (128111 8205 9794 65039)
             (128111 8205 9792 65039) (128110 127999 8205 9794 65039)
             (128110 127999 8205 9792 65039) (128110 127998 8205 9794 65039)
             (128110 127998 8205 9792 65039) (128110 127997 8205 9794 65039)
             (128110 127997 8205 9792 65039) (128110 127996 8205 9794 65039)
             (128110 127996 8205 9792 65039) (128110 127995 8205 9794 65039)
             (128110 127995 8205 9792 65039) (128110 8205 9794 65039)
             (128110 8205 9792 65039) (127948 65039 8205 9794 65039)
             (127948 65039 8205 9792 65039) (127948 127999 8205 9794 65039)
             (127948 127999 8205 9792 65039) (127948 127998 8205 9794 65039)
             (127948 127998 8205 9792 65039) (127948 127997 8205 9794 65039)
             (127948 127997 8205 9792 65039) (127948 127996 8205 9794 65039)
             (127948 127996 8205 9792 65039) (127948 127995 8205 9794 65039)
             (127948 127995 8205 9792 65039) (127947 65039 8205 9794 65039)
             (127947 65039 8205 9792 65039) (127947 127999 8205 9794 65039)
             (127947 127999 8205 9792 65039) (127947 127998 8205 9794 65039)
             (127947 127998 8205 9792 65039) (127947 127997 8205 9794 65039)
             (127947 127997 8205 9792 65039) (127947 127996 8205 9794 65039)
             (127947 127996 8205 9792 65039) (127947 127995 8205 9794 65039)
             (127947 127995 8205 9792 65039) (127946 127999 8205 9794 65039)
             (127946 127999 8205 9792 65039) (127946 127998 8205 9794 65039)
             (127946 127998 8205 9792 65039) (127946 127997 8205 9794 65039)
             (127946 127997 8205 9792 65039) (127946 127996 8205 9794 65039)
             (127946 127996 8205 9792 65039) (127946 127995 8205 9794 65039)
             (127946 127995 8205 9792 65039) (127946 8205 9794 65039)
             (127946 8205 9792 65039) (127940 127999 8205 9794 65039)
             (127940 127999 8205 9792 65039) (127940 127998 8205 9794 65039)
             (127940 127998 8205 9792 65039) (127940 127997 8205 9794 65039)
             (127940 127997 8205 9792 65039) (127940 127996 8205 9794 65039)
             (127940 127996 8205 9792 65039) (127940 127995 8205 9794 65039)
             (127940 127995 8205 9792 65039) (127940 8205 9794 65039)
             (127940 8205 9792 65039) (127939 127999 8205 9794 65039)
             (127939 127999 8205 9792 65039) (127939 127998 8205 9794 65039)
             (127939 127998 8205 9792 65039) (127939 127997 8205 9794 65039)
             (127939 127997 8205 9792 65039) (127939 127996 8205 9794 65039)
             (127939 127996 8205 9792 65039) (127939 127995 8205 9794 65039)
             (127939 127995 8205 9792 65039) (127939 8205 9794 65039)
             (127939 8205 9792 65039) (9977 65039 8205 9794 65039)
             (9977 65039 8205 9792 65039) (9977 127999 8205 9794 65039)
             (9977 127999 8205 9792 65039) (9977 127998 8205 9794 65039)
             (9977 127998 8205 9792 65039) (9977 127997 8205 9794 65039)
             (9977 127997 8205 9792 65039) (9977 127996 8205 9794 65039)
             (9977 127996 8205 9792 65039) (9977 127995 8205 9794 65039)
             (9977 127995 8205 9792 65039) (129489 127999 8205 129469)
             (129489 127999 8205 129468) (129489 127999 8205 129455)
             (129489 127999 8205 128658) (129489 127999 8205 128640)
             (129489 127999 8205 128300) (129489 127999 8205 128295)
             (129489 127999 8205 128188) (129489 127999 8205 128187)
             (129489 127999 8205 127981) (129489 127999 8205 127979)
             (129489 127999 8205 127912) (129489 127999 8205 127908)
             (129489 127999 8205 127891) (129489 127999 8205 127868)
             (129489 127999 8205 127859) (129489 127999 8205 127806)
             (129489 127999 8205 9992 65039) (129489 127999 8205 9878 65039)
             (129489 127999 8205 9877 65039) (129489 127998 8205 129469)
             (129489 127998 8205 129468) (129489 127998 8205 129455)
             (129489 127998 8205 128658) (129489 127998 8205 128640)
             (129489 127998 8205 128300) (129489 127998 8205 128295)
             (129489 127998 8205 128188) (129489 127998 8205 128187)
             (129489 127998 8205 127981) (129489 127998 8205 127979)
             (129489 127998 8205 127912) (129489 127998 8205 127908)
             (129489 127998 8205 127891) (129489 127998 8205 127868)
             (129489 127998 8205 127859) (129489 127998 8205 127806)
             (129489 127998 8205 9992 65039) (129489 127998 8205 9878 65039)
             (129489 127998 8205 9877 65039) (129489 127997 8205 129469)
             (129489 127997 8205 129468) (129489 127997 8205 129455)
             (129489 127997 8205 128658) (129489 127997 8205 128640)
             (129489 127997 8205 128300) (129489 127997 8205 128295)
             (129489 127997 8205 128188) (129489 127997 8205 128187)
             (129489 127997 8205 127981) (129489 127997 8205 127979)
             (129489 127997 8205 127912) (129489 127997 8205 127908)
             (129489 127997 8205 127891) (129489 127997 8205 127868)
             (129489 127997 8205 127859) (129489 127997 8205 127806)
             (129489 127997 8205 9992 65039) (129489 127997 8205 9878 65039)
             (129489 127997 8205 9877 65039) (129489 127996 8205 129469)
             (129489 127996 8205 129468) (129489 127996 8205 129455)
             (129489 127996 8205 128658) (129489 127996 8205 128640)
             (129489 127996 8205 128300) (129489 127996 8205 128295)
             (129489 127996 8205 128188) (129489 127996 8205 128187)
             (129489 127996 8205 127981) (129489 127996 8205 127979)
             (129489 127996 8205 127912) (129489 127996 8205 127908)
             (129489 127996 8205 127891) (129489 127996 8205 127868)
             (129489 127996 8205 127859) (129489 127996 8205 127806)
             (129489 127996 8205 9992 65039) (129489 127996 8205 9878 65039)
             (129489 127996 8205 9877 65039) (129489 127995 8205 129469)
             (129489 127995 8205 129468) (129489 127995 8205 129455)
             (129489 127995 8205 128658) (129489 127995 8205 128640)
             (129489 127995 8205 128300) (129489 127995 8205 128295)
             (129489 127995 8205 128188) (129489 127995 8205 128187)
             (129489 127995 8205 127981) (129489 127995 8205 127979)
             (129489 127995 8205 127912) (129489 127995 8205 127908)
             (129489 127995 8205 127891) (129489 127995 8205 127868)
             (129489 127995 8205 127859) (129489 127995 8205 127806)
             (129489 127995 8205 9992 65039) (129489 127995 8205 9878 65039)
             (129489 127995 8205 9877 65039) (129489 8205 129469)
             (129489 8205 129468) (129489 8205 129455) (129489 8205 128658)
             (129489 8205 128640) (129489 8205 128300) (129489 8205 128295)
             (129489 8205 128188) (129489 8205 128187) (129489 8205 127981)
             (129489 8205 127979) (129489 8205 127912) (129489 8205 127908)
             (129489 8205 127891) (129489 8205 127868) (129489 8205 127859)
             (129489 8205 127806) (129489 8205 9992 65039)
             (129489 8205 9878 65039) (129489 8205 9877 65039)
             (128105 127999 8205 129469) (128105 127999 8205 129468)
             (128105 127999 8205 129455) (128105 127999 8205 128658)
             (128105 127999 8205 128640) (128105 127999 8205 128300)
             (128105 127999 8205 128295) (128105 127999 8205 128188)
             (128105 127999 8205 128187) (128105 127999 8205 127981)
             (128105 127999 8205 127979) (128105 127999 8205 127912)
             (128105 127999 8205 127908) (128105 127999 8205 127891)
             (128105 127999 8205 127868) (128105 127999 8205 127859)
             (128105 127999 8205 127806) (128105 127999 8205 9992 65039)
             (128105 127999 8205 9878 65039) (128105 127999 8205 9877 65039)
             (128105 127998 8205 129469) (128105 127998 8205 129468)
             (128105 127998 8205 129455) (128105 127998 8205 128658)
             (128105 127998 8205 128640) (128105 127998 8205 128300)
             (128105 127998 8205 128295) (128105 127998 8205 128188)
             (128105 127998 8205 128187) (128105 127998 8205 127981)
             (128105 127998 8205 127979) (128105 127998 8205 127912)
             (128105 127998 8205 127908) (128105 127998 8205 127891)
             (128105 127998 8205 127868) (128105 127998 8205 127859)
             (128105 127998 8205 127806) (128105 127998 8205 9992 65039)
             (128105 127998 8205 9878 65039) (128105 127998 8205 9877 65039)
             (128105 127997 8205 129469) (128105 127997 8205 129468)
             (128105 127997 8205 129455) (128105 127997 8205 128658)
             (128105 127997 8205 128640) (128105 127997 8205 128300)
             (128105 127997 8205 128295) (128105 127997 8205 128188)
             (128105 127997 8205 128187) (128105 127997 8205 127981)
             (128105 127997 8205 127979) (128105 127997 8205 127912)
             (128105 127997 8205 127908) (128105 127997 8205 127891)
             (128105 127997 8205 127868) (128105 127997 8205 127859)
             (128105 127997 8205 127806) (128105 127997 8205 9992 65039)
             (128105 127997 8205 9878 65039) (128105 127997 8205 9877 65039)
             (128105 127996 8205 129469) (128105 127996 8205 129468)
             (128105 127996 8205 129455) (128105 127996 8205 128658)
             (128105 127996 8205 128640) (128105 127996 8205 128300)
             (128105 127996 8205 128295) (128105 127996 8205 128188)
             (128105 127996 8205 128187) (128105 127996 8205 127981)
             (128105 127996 8205 127979) (128105 127996 8205 127912)
             (128105 127996 8205 127908) (128105 127996 8205 127891)
             (128105 127996 8205 127868) (128105 127996 8205 127859)
             (128105 127996 8205 127806) (128105 127996 8205 9992 65039)
             (128105 127996 8205 9878 65039) (128105 127996 8205 9877 65039)
             (128105 127995 8205 129469) (128105 127995 8205 129468)
             (128105 127995 8205 129455) (128105 127995 8205 128658)
             (128105 127995 8205 128640) (128105 127995 8205 128300)
             (128105 127995 8205 128295) (128105 127995 8205 128188)
             (128105 127995 8205 128187) (128105 127995 8205 127981)
             (128105 127995 8205 127979) (128105 127995 8205 127912)
             (128105 127995 8205 127908) (128105 127995 8205 127891)
             (128105 127995 8205 127868) (128105 127995 8205 127859)
             (128105 127995 8205 127806) (128105 127995 8205 9992 65039)
             (128105 127995 8205 9878 65039) (128105 127995 8205 9877 65039)
             (128105 8205 129469) (128105 8205 129468) (128105 8205 129455)
             (128105 8205 128658) (128105 8205 128640) (128105 8205 128300)
             (128105 8205 128295) (128105 8205 128188) (128105 8205 128187)
             (128105 8205 127981) (128105 8205 127979) (128105 8205 127912)
             (128105 8205 127908) (128105 8205 127891) (128105 8205 127868)
             (128105 8205 127859) (128105 8205 127806) (128105 8205 9992 65039)
             (128105 8205 9878 65039) (128105 8205 9877 65039)
             (128104 127999 8205 129469) (128104 127999 8205 129468)
             (128104 127999 8205 129455) (128104 127999 8205 128658)
             (128104 127999 8205 128640) (128104 127999 8205 128300)
             (128104 127999 8205 128295) (128104 127999 8205 128188)
             (128104 127999 8205 128187) (128104 127999 8205 127981)
             (128104 127999 8205 127979) (128104 127999 8205 127912)
             (128104 127999 8205 127908) (128104 127999 8205 127891)
             (128104 127999 8205 127868) (128104 127999 8205 127859)
             (128104 127999 8205 127806) (128104 127999 8205 9992 65039)
             (128104 127999 8205 9878 65039) (128104 127999 8205 9877 65039)
             (128104 127998 8205 129469) (128104 127998 8205 129468)
             (128104 127998 8205 129455) (128104 127998 8205 128658)
             (128104 127998 8205 128640) (128104 127998 8205 128300)
             (128104 127998 8205 128295) (128104 127998 8205 128188)
             (128104 127998 8205 128187) (128104 127998 8205 127981)
             (128104 127998 8205 127979) (128104 127998 8205 127912)
             (128104 127998 8205 127908) (128104 127998 8205 127891)
             (128104 127998 8205 127868) (128104 127998 8205 127859)
             (128104 127998 8205 127806) (128104 127998 8205 9992 65039)
             (128104 127998 8205 9878 65039) (128104 127998 8205 9877 65039)
             (128104 127997 8205 129469) (128104 127997 8205 129468)
             (128104 127997 8205 129455) (128104 127997 8205 128658)
             (128104 127997 8205 128640) (128104 127997 8205 128300)
             (128104 127997 8205 128295) (128104 127997 8205 128188)
             (128104 127997 8205 128187) (128104 127997 8205 127981)
             (128104 127997 8205 127979) (128104 127997 8205 127912)
             (128104 127997 8205 127908) (128104 127997 8205 127891)
             (128104 127997 8205 127868) (128104 127997 8205 127859)
             (128104 127997 8205 127806) (128104 127997 8205 9992 65039)
             (128104 127997 8205 9878 65039) (128104 127997 8205 9877 65039)
             (128104 127996 8205 129469) (128104 127996 8205 129468)
             (128104 127996 8205 129455) (128104 127996 8205 128658)
             (128104 127996 8205 128640) (128104 127996 8205 128300)
             (128104 127996 8205 128295) (128104 127996 8205 128188)
             (128104 127996 8205 128187) (128104 127996 8205 127981)
             (128104 127996 8205 127979) (128104 127996 8205 127912)
             (128104 127996 8205 127908) (128104 127996 8205 127891)
             (128104 127996 8205 127868) (128104 127996 8205 127859)
             (128104 127996 8205 127806) (128104 127996 8205 9992 65039)
             (128104 127996 8205 9878 65039) (128104 127996 8205 9877 65039)
             (128104 127995 8205 129469) (128104 127995 8205 129468)
             (128104 127995 8205 129455) (128104 127995 8205 128658)
             (128104 127995 8205 128640) (128104 127995 8205 128300)
             (128104 127995 8205 128295) (128104 127995 8205 128188)
             (128104 127995 8205 128187) (128104 127995 8205 127981)
             (128104 127995 8205 127979) (128104 127995 8205 127912)
             (128104 127995 8205 127908) (128104 127995 8205 127891)
             (128104 127995 8205 127868) (128104 127995 8205 127859)
             (128104 127995 8205 127806) (128104 127995 8205 9992 65039)
             (128104 127995 8205 9878 65039) (128104 127995 8205 9877 65039)
             (128104 8205 129469) (128104 8205 129468) (128104 8205 129455)
             (128104 8205 128658) (128104 8205 128640) (128104 8205 128300)
             (128104 8205 128295) (128104 8205 128188) (128104 8205 128187)
             (128104 8205 127981) (128104 8205 127979) (128104 8205 127912)
             (128104 8205 127908) (128104 8205 127891) (128104 8205 127868)
             (128104 8205 127859) (128104 8205 127806) (128104 8205 9992 65039)
             (128104 8205 9878 65039) (128104 8205 9877 65039)
             (129489 127999 8205 129309 8205 129489 127999)
             (129489 127999 8205 129309 8205 129489 127998)
             (129489 127999 8205 129309 8205 129489 127997)
             (129489 127999 8205 129309 8205 129489 127996)
             (129489 127999 8205 129309 8205 129489 127995)
             (129489 127999 8205 127876)
             (129489 127999 8205 10084 65039 8205 129489 127998)
             (129489 127999 8205 10084 65039 8205 129489 127997)
             (129489 127999 8205 10084 65039 8205 129489 127996)
             (129489 127999 8205 10084 65039 8205 129489 127995)
             (129489 127999 8205 10084 65039 8205 128139 8205 129489 127998)
             (129489 127999 8205 10084 65039 8205 128139 8205 129489 127997)
             (129489 127999 8205 10084 65039 8205 128139 8205 129489 127996)
             (129489 127999 8205 10084 65039 8205 128139 8205 129489 127995)
             (129489 127998 8205 129309 8205 129489 127999)
             (129489 127998 8205 129309 8205 129489 127998)
             (129489 127998 8205 129309 8205 129489 127997)
             (129489 127998 8205 129309 8205 129489 127996)
             (129489 127998 8205 129309 8205 129489 127995)
             (129489 127998 8205 127876)
             (129489 127998 8205 10084 65039 8205 129489 127999)
             (129489 127998 8205 10084 65039 8205 129489 127997)
             (129489 127998 8205 10084 65039 8205 129489 127996)
             (129489 127998 8205 10084 65039 8205 129489 127995)
             (129489 127998 8205 10084 65039 8205 128139 8205 129489 127999)
             (129489 127998 8205 10084 65039 8205 128139 8205 129489 127997)
             (129489 127998 8205 10084 65039 8205 128139 8205 129489 127996)
             (129489 127998 8205 10084 65039 8205 128139 8205 129489 127995)
             (129489 127997 8205 129309 8205 129489 127999)
             (129489 127997 8205 129309 8205 129489 127998)
             (129489 127997 8205 129309 8205 129489 127997)
             (129489 127997 8205 129309 8205 129489 127996)
             (129489 127997 8205 129309 8205 129489 127995)
             (129489 127997 8205 127876)
             (129489 127997 8205 10084 65039 8205 129489 127999)
             (129489 127997 8205 10084 65039 8205 129489 127998)
             (129489 127997 8205 10084 65039 8205 129489 127996)
             (129489 127997 8205 10084 65039 8205 129489 127995)
             (129489 127997 8205 10084 65039 8205 128139 8205 129489 127999)
             (129489 127997 8205 10084 65039 8205 128139 8205 129489 127998)
             (129489 127997 8205 10084 65039 8205 128139 8205 129489 127996)
             (129489 127997 8205 10084 65039 8205 128139 8205 129489 127995)
             (129489 127996 8205 129309 8205 129489 127999)
             (129489 127996 8205 129309 8205 129489 127998)
             (129489 127996 8205 129309 8205 129489 127997)
             (129489 127996 8205 129309 8205 129489 127996)
             (129489 127996 8205 129309 8205 129489 127995)
             (129489 127996 8205 127876)
             (129489 127996 8205 10084 65039 8205 129489 127999)
             (129489 127996 8205 10084 65039 8205 129489 127998)
             (129489 127996 8205 10084 65039 8205 129489 127997)
             (129489 127996 8205 10084 65039 8205 129489 127995)
             (129489 127996 8205 10084 65039 8205 128139 8205 129489 127999)
             (129489 127996 8205 10084 65039 8205 128139 8205 129489 127998)
             (129489 127996 8205 10084 65039 8205 128139 8205 129489 127997)
             (129489 127996 8205 10084 65039 8205 128139 8205 129489 127995)
             (129489 127995 8205 129309 8205 129489 127999)
             (129489 127995 8205 129309 8205 129489 127998)
             (129489 127995 8205 129309 8205 129489 127997)
             (129489 127995 8205 129309 8205 129489 127996)
             (129489 127995 8205 129309 8205 129489 127995)
             (129489 127995 8205 127876)
             (129489 127995 8205 10084 65039 8205 129489 127999)
             (129489 127995 8205 10084 65039 8205 129489 127998)
             (129489 127995 8205 10084 65039 8205 129489 127997)
             (129489 127995 8205 10084 65039 8205 129489 127996)
             (129489 127995 8205 10084 65039 8205 128139 8205 129489 127999)
             (129489 127995 8205 10084 65039 8205 128139 8205 129489 127998)
             (129489 127995 8205 10084 65039 8205 128139 8205 129489 127997)
             (129489 127995 8205 10084 65039 8205 128139 8205 129489 127996)
             (129489 8205 129309 8205 129489)
             (128105 127999 8205 129309 8205 128105 127998)
             (128105 127999 8205 129309 8205 128105 127997)
             (128105 127999 8205 129309 8205 128105 127996)
             (128105 127999 8205 129309 8205 128105 127995)
             (128105 127999 8205 129309 8205 128104 127998)
             (128105 127999 8205 129309 8205 128104 127997)
             (128105 127999 8205 129309 8205 128104 127996)
             (128105 127999 8205 129309 8205 128104 127995)
             (128105 127999 8205 10084 65039 8205 128139 8205 128105 127999)
             (128105 127999 8205 10084 65039 8205 128139 8205 128105 127998)
             (128105 127999 8205 10084 65039 8205 128139 8205 128105 127997)
             (128105 127999 8205 10084 65039 8205 128139 8205 128105 127996)
             (128105 127999 8205 10084 65039 8205 128139 8205 128105 127995)
             (128105 127999 8205 10084 65039 8205 128139 8205 128104 127999)
             (128105 127999 8205 10084 65039 8205 128139 8205 128104 127998)
             (128105 127999 8205 10084 65039 8205 128139 8205 128104 127997)
             (128105 127999 8205 10084 65039 8205 128139 8205 128104 127996)
             (128105 127999 8205 10084 65039 8205 128139 8205 128104 127995)
             (128105 127999 8205 10084 65039 8205 128105 127999)
             (128105 127999 8205 10084 65039 8205 128105 127998)
             (128105 127999 8205 10084 65039 8205 128105 127997)
             (128105 127999 8205 10084 65039 8205 128105 127996)
             (128105 127999 8205 10084 65039 8205 128105 127995)
             (128105 127999 8205 10084 65039 8205 128104 127999)
             (128105 127999 8205 10084 65039 8205 128104 127998)
             (128105 127999 8205 10084 65039 8205 128104 127997)
             (128105 127999 8205 10084 65039 8205 128104 127996)
             (128105 127999 8205 10084 65039 8205 128104 127995)
             (128105 127998 8205 129309 8205 128105 127999)
             (128105 127998 8205 129309 8205 128105 127997)
             (128105 127998 8205 129309 8205 128105 127996)
             (128105 127998 8205 129309 8205 128105 127995)
             (128105 127998 8205 129309 8205 128104 127999)
             (128105 127998 8205 129309 8205 128104 127997)
             (128105 127998 8205 129309 8205 128104 127996)
             (128105 127998 8205 129309 8205 128104 127995)
             (128105 127998 8205 10084 65039 8205 128139 8205 128105 127999)
             (128105 127998 8205 10084 65039 8205 128139 8205 128105 127998)
             (128105 127998 8205 10084 65039 8205 128139 8205 128105 127997)
             (128105 127998 8205 10084 65039 8205 128139 8205 128105 127996)
             (128105 127998 8205 10084 65039 8205 128139 8205 128105 127995)
             (128105 127998 8205 10084 65039 8205 128139 8205 128104 127999)
             (128105 127998 8205 10084 65039 8205 128139 8205 128104 127998)
             (128105 127998 8205 10084 65039 8205 128139 8205 128104 127997)
             (128105 127998 8205 10084 65039 8205 128139 8205 128104 127996)
             (128105 127998 8205 10084 65039 8205 128139 8205 128104 127995)
             (128105 127998 8205 10084 65039 8205 128105 127999)
             (128105 127998 8205 10084 65039 8205 128105 127998)
             (128105 127998 8205 10084 65039 8205 128105 127997)
             (128105 127998 8205 10084 65039 8205 128105 127996)
             (128105 127998 8205 10084 65039 8205 128105 127995)
             (128105 127998 8205 10084 65039 8205 128104 127999)
             (128105 127998 8205 10084 65039 8205 128104 127998)
             (128105 127998 8205 10084 65039 8205 128104 127997)
             (128105 127998 8205 10084 65039 8205 128104 127996)
             (128105 127998 8205 10084 65039 8205 128104 127995)
             (128105 127997 8205 129309 8205 128105 127999)
             (128105 127997 8205 129309 8205 128105 127998)
             (128105 127997 8205 129309 8205 128105 127996)
             (128105 127997 8205 129309 8205 128105 127995)
             (128105 127997 8205 129309 8205 128104 127999)
             (128105 127997 8205 129309 8205 128104 127998)
             (128105 127997 8205 129309 8205 128104 127996)
             (128105 127997 8205 129309 8205 128104 127995)
             (128105 127997 8205 10084 65039 8205 128139 8205 128105 127999)
             (128105 127997 8205 10084 65039 8205 128139 8205 128105 127998)
             (128105 127997 8205 10084 65039 8205 128139 8205 128105 127997)
             (128105 127997 8205 10084 65039 8205 128139 8205 128105 127996)
             (128105 127997 8205 10084 65039 8205 128139 8205 128105 127995)
             (128105 127997 8205 10084 65039 8205 128139 8205 128104 127999)
             (128105 127997 8205 10084 65039 8205 128139 8205 128104 127998)
             (128105 127997 8205 10084 65039 8205 128139 8205 128104 127997)
             (128105 127997 8205 10084 65039 8205 128139 8205 128104 127996)
             (128105 127997 8205 10084 65039 8205 128139 8205 128104 127995)
             (128105 127997 8205 10084 65039 8205 128105 127999)
             (128105 127997 8205 10084 65039 8205 128105 127998)
             (128105 127997 8205 10084 65039 8205 128105 127997)
             (128105 127997 8205 10084 65039 8205 128105 127996)
             (128105 127997 8205 10084 65039 8205 128105 127995)
             (128105 127997 8205 10084 65039 8205 128104 127999)
             (128105 127997 8205 10084 65039 8205 128104 127998)
             (128105 127997 8205 10084 65039 8205 128104 127997)
             (128105 127997 8205 10084 65039 8205 128104 127996)
             (128105 127997 8205 10084 65039 8205 128104 127995)
             (128105 127996 8205 129309 8205 128105 127999)
             (128105 127996 8205 129309 8205 128105 127998)
             (128105 127996 8205 129309 8205 128105 127997)
             (128105 127996 8205 129309 8205 128105 127995)
             (128105 127996 8205 129309 8205 128104 127999)
             (128105 127996 8205 129309 8205 128104 127998)
             (128105 127996 8205 129309 8205 128104 127997)
             (128105 127996 8205 129309 8205 128104 127995)
             (128105 127996 8205 10084 65039 8205 128139 8205 128105 127999)
             (128105 127996 8205 10084 65039 8205 128139 8205 128105 127998)
             (128105 127996 8205 10084 65039 8205 128139 8205 128105 127997)
             (128105 127996 8205 10084 65039 8205 128139 8205 128105 127996)
             (128105 127996 8205 10084 65039 8205 128139 8205 128105 127995)
             (128105 127996 8205 10084 65039 8205 128139 8205 128104 127999)
             (128105 127996 8205 10084 65039 8205 128139 8205 128104 127998)
             (128105 127996 8205 10084 65039 8205 128139 8205 128104 127997)
             (128105 127996 8205 10084 65039 8205 128139 8205 128104 127996)
             (128105 127996 8205 10084 65039 8205 128139 8205 128104 127995)
             (128105 127996 8205 10084 65039 8205 128105 127999)
             (128105 127996 8205 10084 65039 8205 128105 127998)
             (128105 127996 8205 10084 65039 8205 128105 127997)
             (128105 127996 8205 10084 65039 8205 128105 127996)
             (128105 127996 8205 10084 65039 8205 128105 127995)
             (128105 127996 8205 10084 65039 8205 128104 127999)
             (128105 127996 8205 10084 65039 8205 128104 127998)
             (128105 127996 8205 10084 65039 8205 128104 127997)
             (128105 127996 8205 10084 65039 8205 128104 127996)
             (128105 127996 8205 10084 65039 8205 128104 127995)
             (128105 127995 8205 129309 8205 128105 127999)
             (128105 127995 8205 129309 8205 128105 127998)
             (128105 127995 8205 129309 8205 128105 127997)
             (128105 127995 8205 129309 8205 128105 127996)
             (128105 127995 8205 129309 8205 128104 127999)
             (128105 127995 8205 129309 8205 128104 127998)
             (128105 127995 8205 129309 8205 128104 127997)
             (128105 127995 8205 129309 8205 128104 127996)
             (128105 127995 8205 10084 65039 8205 128139 8205 128105 127999)
             (128105 127995 8205 10084 65039 8205 128139 8205 128105 127998)
             (128105 127995 8205 10084 65039 8205 128139 8205 128105 127997)
             (128105 127995 8205 10084 65039 8205 128139 8205 128105 127996)
             (128105 127995 8205 10084 65039 8205 128139 8205 128105 127995)
             (128105 127995 8205 10084 65039 8205 128139 8205 128104 127999)
             (128105 127995 8205 10084 65039 8205 128139 8205 128104 127998)
             (128105 127995 8205 10084 65039 8205 128139 8205 128104 127997)
             (128105 127995 8205 10084 65039 8205 128139 8205 128104 127996)
             (128105 127995 8205 10084 65039 8205 128139 8205 128104 127995)
             (128105 127995 8205 10084 65039 8205 128105 127999)
             (128105 127995 8205 10084 65039 8205 128105 127998)
             (128105 127995 8205 10084 65039 8205 128105 127997)
             (128105 127995 8205 10084 65039 8205 128105 127996)
             (128105 127995 8205 10084 65039 8205 128105 127995)
             (128105 127995 8205 10084 65039 8205 128104 127999)
             (128105 127995 8205 10084 65039 8205 128104 127998)
             (128105 127995 8205 10084 65039 8205 128104 127997)
             (128105 127995 8205 10084 65039 8205 128104 127996)
             (128105 127995 8205 10084 65039 8205 128104 127995)
             (128105 8205 128105 8205 128103 8205 128103)
             (128105 8205 128105 8205 128103 8205 128102)
             (128105 8205 128105 8205 128103)
             (128105 8205 128105 8205 128102 8205 128102)
             (128105 8205 128105 8205 128102) (128105 8205 128103 8205 128103)
             (128105 8205 128103 8205 128102) (128105 8205 128103)
             (128105 8205 128102 8205 128102) (128105 8205 128102)
             (128105 8205 10084 65039 8205 128139 8205 128105)
             (128105 8205 10084 65039 8205 128139 8205 128104)
             (128105 8205 10084 65039 8205 128105)
             (128105 8205 10084 65039 8205 128104)
             (128104 127999 8205 129309 8205 128104 127998)
             (128104 127999 8205 129309 8205 128104 127997)
             (128104 127999 8205 129309 8205 128104 127996)
             (128104 127999 8205 129309 8205 128104 127995)
             (128104 127999 8205 10084 65039 8205 128139 8205 128104 127999)
             (128104 127999 8205 10084 65039 8205 128139 8205 128104 127998)
             (128104 127999 8205 10084 65039 8205 128139 8205 128104 127997)
             (128104 127999 8205 10084 65039 8205 128139 8205 128104 127996)
             (128104 127999 8205 10084 65039 8205 128139 8205 128104 127995)
             (128104 127999 8205 10084 65039 8205 128104 127999)
             (128104 127999 8205 10084 65039 8205 128104 127998)
             (128104 127999 8205 10084 65039 8205 128104 127997)
             (128104 127999 8205 10084 65039 8205 128104 127996)
             (128104 127999 8205 10084 65039 8205 128104 127995)
             (128104 127998 8205 129309 8205 128104 127999)
             (128104 127998 8205 129309 8205 128104 127997)
             (128104 127998 8205 129309 8205 128104 127996)
             (128104 127998 8205 129309 8205 128104 127995)
             (128104 127998 8205 10084 65039 8205 128139 8205 128104 127999)
             (128104 127998 8205 10084 65039 8205 128139 8205 128104 127998)
             (128104 127998 8205 10084 65039 8205 128139 8205 128104 127997)
             (128104 127998 8205 10084 65039 8205 128139 8205 128104 127996)
             (128104 127998 8205 10084 65039 8205 128139 8205 128104 127995)
             (128104 127998 8205 10084 65039 8205 128104 127999)
             (128104 127998 8205 10084 65039 8205 128104 127998)
             (128104 127998 8205 10084 65039 8205 128104 127997)
             (128104 127998 8205 10084 65039 8205 128104 127996)
             (128104 127998 8205 10084 65039 8205 128104 127995)
             (128104 127997 8205 129309 8205 128104 127999)
             (128104 127997 8205 129309 8205 128104 127998)
             (128104 127997 8205 129309 8205 128104 127996)
             (128104 127997 8205 129309 8205 128104 127995)
             (128104 127997 8205 10084 65039 8205 128139 8205 128104 127999)
             (128104 127997 8205 10084 65039 8205 128139 8205 128104 127998)
             (128104 127997 8205 10084 65039 8205 128139 8205 128104 127997)
             (128104 127997 8205 10084 65039 8205 128139 8205 128104 127996)
             (128104 127997 8205 10084 65039 8205 128139 8205 128104 127995)
             (128104 127997 8205 10084 65039 8205 128104 127999)
             (128104 127997 8205 10084 65039 8205 128104 127998)
             (128104 127997 8205 10084 65039 8205 128104 127997)
             (128104 127997 8205 10084 65039 8205 128104 127996)
             (128104 127997 8205 10084 65039 8205 128104 127995)
             (128104 127996 8205 129309 8205 128104 127999)
             (128104 127996 8205 129309 8205 128104 127998)
             (128104 127996 8205 129309 8205 128104 127997)
             (128104 127996 8205 129309 8205 128104 127995)
             (128104 127996 8205 10084 65039 8205 128139 8205 128104 127999)
             (128104 127996 8205 10084 65039 8205 128139 8205 128104 127998)
             (128104 127996 8205 10084 65039 8205 128139 8205 128104 127997)
             (128104 127996 8205 10084 65039 8205 128139 8205 128104 127996)
             (128104 127996 8205 10084 65039 8205 128139 8205 128104 127995)
             (128104 127996 8205 10084 65039 8205 128104 127999)
             (128104 127996 8205 10084 65039 8205 128104 127998)
             (128104 127996 8205 10084 65039 8205 128104 127997)
             (128104 127996 8205 10084 65039 8205 128104 127996)
             (128104 127996 8205 10084 65039 8205 128104 127995)
             (128104 127995 8205 129309 8205 128104 127999)
             (128104 127995 8205 129309 8205 128104 127998)
             (128104 127995 8205 129309 8205 128104 127997)
             (128104 127995 8205 129309 8205 128104 127996)
             (128104 127995 8205 10084 65039 8205 128139 8205 128104 127999)
             (128104 127995 8205 10084 65039 8205 128139 8205 128104 127998)
             (128104 127995 8205 10084 65039 8205 128139 8205 128104 127997)
             (128104 127995 8205 10084 65039 8205 128139 8205 128104 127996)
             (128104 127995 8205 10084 65039 8205 128139 8205 128104 127995)
             (128104 127995 8205 10084 65039 8205 128104 127999)
             (128104 127995 8205 10084 65039 8205 128104 127998)
             (128104 127995 8205 10084 65039 8205 128104 127997)
             (128104 127995 8205 10084 65039 8205 128104 127996)
             (128104 127995 8205 10084 65039 8205 128104 127995)
             (128104 8205 128105 8205 128103 8205 128103)
             (128104 8205 128105 8205 128103 8205 128102)
             (128104 8205 128105 8205 128103)
             (128104 8205 128105 8205 128102 8205 128102)
             (128104 8205 128105 8205 128102)
             (128104 8205 128104 8205 128103 8205 128103)
             (128104 8205 128104 8205 128103 8205 128102)
             (128104 8205 128104 8205 128103)
             (128104 8205 128104 8205 128102 8205 128102)
             (128104 8205 128104 8205 128102) (128104 8205 128103 8205 128103)
             (128104 8205 128103 8205 128102) (128104 8205 128103)
             (128104 8205 128102 8205 128102) (128104 8205 128102)
             (128104 8205 10084 65039 8205 128139 8205 128104)
             (128104 8205 10084 65039 8205 128104))))

(defgeneric emojip (codepoints-chars))

(defmethod emojip ((codepoints-chars list))
  (or (emoji-zwj-sequences-p codepoints-chars)
      (emoji-sequences-p     codepoints-chars)))

(defgeneric starting-emoji (object))

(defmethod starting-emoji ((object list))
    (loop for end from 1 to +max-codepoints-emoji-lengths+
          when (>= (length object) end) do
            (alexandria:when-let* ((sample     (subseq object 0 end))
                                   (emoji-list (emojip sample)))
              (return-from starting-emoji emoji-list)))
    nil)

(defmethod starting-emoji ((object string))
  (let ((chars-list (coerce object 'list)))
    (starting-emoji chars-list)))
